/**
This is pseudocode (algorithms) for DynamicArray.

It is for "ideas".

It uses an iterator. You do not need to use an
iterator, you can use other iteration techniques
if you prefer, e.g. for loops, etc.
*/

class DynamicArray {

	//----------------------------------------------------------
	//ivars

    //fixedArray                                -- primary ivar
    //_size, growthFactor, initialCapacity      -- support/administation ivars

	//----------------------------------------------------------
	//construction

	//----------------------------------------------------------
	// Pubic Instance Methods -- for DynamicList spec - Core

    size() {
        return this._size;
    }

    isEmpty() {
        return this.size() == 0;
    }

	//----------------------------------------------------------
	// Convenience Methods

    toString() {
        var max, postfix, printables;
        max = 4;
        printables = [];
        if (this.size() > max) {
            printables = this.subList(0, max + 1).toArray();
            postfix = `... (size={$this.size()})`;
        }
        else
            printables = this.toArray();
        return `[${printables.toString()}]`;
        }

    forEach(actionFct) {
        var iter = this.iterator();
        while (iter.hasNext())
            actionFct(iter.next());
    }

    toArray() {
        var array, index, iter;
        array = Array(this.size());
        iter = this.iterator();
        index = 0;
        while (iter.hasNext())
            array[index++] = iter.next();
        return array;
   }

	//----------------------------------------------------------
	// Accessing Elements

    get(index) {
        if (!this.checkIndex(index))
            throw 'get index out of bounds: ' + index;
        return this.fixedArray[index];
    }

    first() {
        if (this.isEmpty())
            throw '(first) Attempted to access element in empty list';
        return this.get(0);
    }

    findFirst(searchFct) {
        /*
        for (let i = 0; i < this.size(); i++)
            if (searchFct(this.get(i)))
                return i;
        return -1;
        */
        var iter, index;
        iter = this.iterator();
        index = 0;
        while (iter.hasNext()) {
            if (searchFct(iter.next()))
                return index
            index++;
        }
        return -1;
    }

	//----------------------------------------------------------
	// Pubic Instance Methods -- for DynamicList spec - Changing/Mutating

    add(newElement) {
        this.addLast(newElement);
    }

    addLast(newElement) {
        this.checkAndAdjustCapacity(1);
        this.fixedArray[this.lastIndex() + 1] = newElement;
        this._size++;
    }

	//----------------------------------------------------------
	// Utility Methods

    select(selectFct) {
        var newElements, iter;
        newElements = DynamicArray.newEmpty();
        iter = this.iterator();
        while (iter.hasNext()) {
            if (selectFct(iter.peek()))
                newElements.add(iter.peek())
            iter.next();
        }
        return newElements;
    }

	//----------------------------------------------------------
	// Optional Methods

    iterator() {
        return new DynArrayIterator(this);
    }


	//----------------------------------------------------------
	// Helper Methods (Instance)

    initialize() {
        this.fixedArray = this.constructor.newFixedArray(this.initialCapacity);
        this._size = 0;
    }

    checkAndAdjustCapacity(number) {
        //TODO
    }

    lastIndex() {
        return this.size() - 1;
    }

    checkIndex(index) {
        return index >= 0 && index < this.size();
    }

	//----------------------------------------------------------
	// Helper Methods (Static)

    static defaultGrowthFactor() {
        return 2;
    }

    static computeSize(size, growthFactor) {
        return Math.round(size * growthFactor);
    }

    static newFixedArray(aSize) {
        //Return new fixed array (fa) that has size per the passed input arg "aSize"
        //The "length" property of the new array will be "aSize"
        let fa = Array(aSize);
        // Fill with nulls
        fa.fill(null);
        // Assure it is fixed
        Object.seal(fa);
        return fa;
    }

	//----------------------------------------------------------

    subList(start, stop) {
		//TODO
    }

	//----------------------------------------------------------
	//construction

    static from(elements) {
        var gf, dynArray;
        gf = this.defaultGrowthFactor();
        dynArray = new DynamicArray(gf, Math.max(this.computeSize(elements.length, gf), 10));
        for (let next of elements)
            dynArray.add(next);
        return dynArray;
    }

    static newEmpty() {
        return this.from([]);
    }

    constructor(aGrowthFactor, anInitialCapacity) {
        this.growthFactor = aGrowthFactor;
        this.initialCapacity = anInitialCapacity;
        this.initialize();
    }

}

//--------------------------------------------------------------

class DynArrayIterator {
    //Ivars
    //list, nextIndex
    constructor(list) {
        this.list = list;
        this.nextIndex = 0;
    }

	/** Returns true if we are not at end */
    hasNext() {
		//TODO
		//hint -- use next index vs list size
    }

	/** Returns next element and moves to next position
	(throws exception if at end) */
    next() {
		//TODO
		//hint -- use hasNext and list "get", and throw exception if needed
    }

	/** Returns next element (does not move to next position)
	(returns null if at end) */
    peek() {
		//TODO
		//hint -- use hasNext and list "get"
    }
}

